#!/usr/bin/env python3

"""
Compute a histogram (default 4096 bins) of the input file and plot it.
The intent is to spot float- or int16 data that has been improperly
processed. Possibly by having been stored as int8 at some point in
the processing.

Many data files have a huge number of zeros, in which case real
samples would hardly be visible at all. The code tries to fix this by
clipping the 5 highest entries. Caveat: If the data set contains
ridiculously large spikes then then all the real data will look like
~0 and will all end up in the same bin. Which is then clipped, leaving
only the garbage visible. This is not a bug per se but can be very
confusing. As an example, look at B-03A-87-LA which has real data
+/- 12 or so, and random samples in the +/- 1e16 range.
"""

import numpy as np
import os, sys, time, argparse
import matplotlib.pyplot as plt
from ..api import ZgyReader, ZgyWriter, SampleDataType, ProgressWithDots
from ..test.utils import SDCredentials
from ..iterator import readall

def find_center(r):
    """
    Crop partial bricks outside or below the survey since they
    would just confuse the statistics.
    If the file is large then crop out an area around the center.
    Try to read the full traces but limit the il and xl.
    """
    size = np.array(r.size, dtype=np.int64)
    cropsize = np.minimum(size, np.array([640,640,12800], dtype=np.int64))
    cropsize = (cropsize//64)*64
    cropoffset = ((size - cropsize)//128)*64
    if np.any(size - cropsize >= 64):
        print("Reading center", tuple(cropsize),
              "offset", tuple(cropoffset),
              "of survey size", tuple(r.size))
    return cropsize, cropoffset

def run_open_file(r, datarange, *, progress = None, bincount = 256, crop = False):
    histogram = np.zeros(bincount, dtype=np.int64)
    (cropsize, cropoffset) = find_center(r) if crop else (None, None)
    for _, _, data in readall(r, dtype=np.float32, cropsize=cropsize, cropoffset=cropoffset, progress=progress):
        histogram += np.histogram(data, bins=bincount, range=datarange)[0]
    return histogram

def run(srcfilename, outfilename, datarange=None, bincount=256, crop=False):
    with ZgyReader(srcfilename, iocontext = SDCredentials()) as r:
        if not datarange:
            datarange = tuple(r.datarange)
            # Make symmetric
            if datarange[0] * datarange[1] < 0:
                lim = max(abs(datarange[0]), abs(datarange[1]))
                datarange = (-lim, lim)
        print("Data type", r.datatype,
              "range", datarange, "codingrange", r.datarange,
              "legacy", r._meta._ih._codingrange,
              "statistical", (r.statistics.min, r.statistics.max))
        if r.datatype == SampleDataType.int8:
            bincount=256
        histogram = run_open_file(r, datarange, bincount=bincount, crop=crop,
                                  progress=ProgressWithDots())
    assert len(histogram) == bincount
    plot_histogram("file: " + srcfilename, histogram)
    plt.savefig(outfilename)

def plot_histogram(title, histogram):
    bincount = len(histogram)
    nz = np.count_nonzero(histogram)
    zz = len(histogram) - nz
    #print(list(sorted(histogram)))
    #clipvalue = list(sorted(histogram))[-nz//10] # at 90-percentile
    clipvalue = list(sorted(histogram))[-5] # clip a fixed number of bins
    # Clipping so a large number of zero or close-to-zero won't cause trouble.
    if False:
        print("raw histogram   ", histogram)
        print("sorted histogram", np.array(sorted(histogram))[-6:])
        print("clip from", list(sorted(histogram))[-1],
              "to", clipvalue, "samples/bin")
    # But: If there are fewer than 5 buckets with data in them, don't bother.
    # This can really happen. If a data set has a single huge spike then only
    # the "spike" bin and the "close to zero" bin will be populated.
    # The histogram will be useless, but at least try to show what happened
    # instead of bombing out with a divide by zero.
    # Example data set: sd://sntc/millet/usgs/zgy/B-09-88-LA.zgy
    # which has a real value range +/- 100000 but spices at -9.1e15
    if clipvalue != 0:
        np.clip(histogram, 0, clipvalue, out=histogram)
        #print("clipped hist at ", clipvalue, ":", histogram)
    else:
        print("WARNING: Fewer than 5 buckets filled. Are there spikes?")

    topvalue = list(sorted(histogram))[-1]
    sumvalue = np.sum(histogram)

    print("{0:.1f}% of the {1} histogram bins are empty.".format(
        100.0*zz/len(histogram),
        len(histogram)))

    if sumvalue == 0:
        print("ERROR: histogram is empty.")
    else:
        print("{0:.1f}% of all samples fall in the same bin.".format(
            100.0*topvalue/sumvalue))

    # The plot only allows displaying 512 distinct values without the plot
    # itself starting to generate artifacts. So we must compress, or zoom,
    # or preferably both.
    factor = max(1, bincount // 512)

    zoomed = histogram[(factor-1)*bincount//(2*factor):(factor+1)*bincount//(2*factor)]
    compressed = np.array([np.sum(histogram[ii:ii+factor]) for ii in range(0, bincount, factor)], dtype=np.int64)

    #plt.figure(num=None, figsize=(20, 10), dpi=80)
    fig, (ax1, ax2) = plt.subplots(2, figsize=(12, 7), dpi=100, facecolor="red", edgecolor="green")
    #ax1.set_facecolor('#dddddd')
    ax1.plot(np.arange(len(zoomed)), zoomed)
    ax2.plot(np.arange(len(compressed)), compressed)
    ax1.title.set_text("Zoomed in {0}x, showing the {1} bins at the center of a {2} bin histogram".format(factor, bincount//factor, bincount))
    ax2.title.set_text("Entire histogram, compressed to {0} bins".format(bincount//factor))
    #x = [111,122,155,192,11,123,120,]
    #y = [3,4,3,5,9,10,23]
    #plt.bar(x,y)
    fig.suptitle(title)

if __name__ == "__main__":
    np.seterr(all='raise')
    parser = argparse.ArgumentParser(description='Compute and show histogram')
    parser.add_argument('input', help='ZGY input cube, local or sd://')
    parser.add_argument('--output', default=None, help='PNG output file')
    parser.add_argument('--datarange', nargs=2, default=None, metavar=('MIN', 'MAX'), type=float, help='Histogram limits')
    parser.add_argument('--bins', nargs=1, default=[4096], type=int, help='Number of bins in histogram')
    parser.add_argument("--crop", action='store_true', help='Limit the number of traces scanned')
    args = parser.parse_args()
    args.datarange = tuple(args.datarange) if args.datarange is not None else None
    args.bins = args.bins[0]
    if not args.output:
        args.output = (args.input.rstrip('/').split('/')[-1]) + "-histogram.png"
    print(args)
    run(args.input, args.output, args.datarange, args.bins, args.crop)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
Low level I/O, abstract layer.

This file contains classes that do low level I/O either to on-prem data
using the regular read and write methods of the OS or to seismic store
using SDAPI. Other back ends can easily be added.

Some work is done to consolidate read requests and to buffer write
requests in the cloud back end. This is needed to get acceptable
performance. This explains why this file is quite large.


    impl.file.Config:
      impl.file.FileConfig(Config):
      impl.file.SDConfig(Config):

        * Details such as user credentials etc. established when the
          file is open. Specific to the backend type.
        * Note that there is currently no way to pass a configuration
          object along with every read and write request. This might
          have been useful for a server type application but would
          require the config parameter to ripple across at least 50
          existing methods. I doubt this would be worth the trouble.

    impl.file.FileADT:
      impl.file.LocalFile(FileADT):
        impl.file.LocalFileOther(LocalFile):
        impl.file.LocalFileLinux(LocalFile):
      impl.file.SeismicStoreFile(FileADT):
      impl.file.SeismicStoreFileDelayedWrite(FileADT):

        * Higher level code should only access the polymorphic FileADT
          base class and the impl.file.FileFactory that creates an
          instance of the desired type.
"""
##@package openzgy.impl.file
#@brief Low level I/O, abstract layer.

import os
import json
from enum import Enum
from contextlib import contextmanager, suppress
from collections import namedtuple

from ..exception import *

try:
    import sdglue as sd
except Exception as ex:
    print("seismic store access is not available:", ex)
    sd = None

class UsageHint(Enum):
    Unknown    = 0x00
    TextFile   = 0x01
    Header     = 0x10
    Data       = 0x20
    Compressed = 0x40,

class Config:

    _deprecation_context_warned = False
    @staticmethod
    def _deprecated_context(attrname):
        if not Config._deprecation_context_warned:
            print('DEPRECATION WARNING: IOContext should be a dict. Found "' +
                  attrname + '" as an attribute.')
            Config._deprecation_context_warned = True

    @staticmethod
    def  _get_string_env(context, attrname, name, default):
        try:
            value = context[attrname]
            name = attrname
        except (TypeError, KeyError):
            if hasattr(context, attrname):
                value = getattr(context, attrname)
                name = attrname
                Config._deprecated_context(attrname)
            else:
                value = os.getenv(name, str(default))
        if default and not value:
            print("WARNING: ${0} cannot be empty.".format(name))
            value = default
        return value

    @staticmethod
    def  _get_numeric_env(context, attrname, name, default, min_value, max_value):
        try:
            value = context[attrname]
            name = attrname
        except (TypeError, KeyError):
            if hasattr(context, attrname):
                value = getattr(context, attrname)
                name = attrname
                Config._deprecated_context(attrname)
            else:
                try:
                    value = int(os.getenv(name, str(default)))
                except ValueError:
                    print("WARNING: badly formed number in ${0} ignored.".format(name))
                    value = default
        if value < min_value or value > max_value:
            print("WARNING: ${0} must be be between {1} and {2}.".format(name, min_value, max_value))
            value = default
        return value

class FileConfig(Config):

    def __init__(self, context):
        """
        Currently the on-prem file reader has no user settable configurfation.
        """
        self.maxsize = 0 # No consolidation of read requests.
        self.maxhole = 0 # No fuzzy consolidation.
        self.aligned = 0 # No additional alignment on reads.
        self.segsize = 0 # No buffering of writes.
        self.threads = 1 # No multi threading.

    @staticmethod
    def _redact(s):
        return s if len(s) < 20 else s[:4] + "..." + s[-4:]

    def dump(self):
        return """FileConfig: No user settable configuration."""

def _checkTokenVariableName():
    """
    Temporarily accept both $OPENZGY_TOKEN and $HACK_SAUTH_TEST_TOKEN as the
    credentials to use. Warn if the latter is set.
    Remove this check and the calls to it after a suitable grace period.
    """
    oldtoken = os.getenv("HACK_SAUTH_TEST_TOKEN")
    newtoken = os.getenv("OPENZGY_TOKEN")
    if oldtoken:
        if not newtoken:
            print("WARNING: $HACK_SAUTH_TEST_TOKEN is deprecated. Please use $OPENZGY_TOKEN instead.")
            os.environ["OPENZGY_TOKEN"] = oldtoken
        elif oldtoken != newtoken:
            print("WARNING: $HACK_SAUTH_TEST_TOKEN is deprecated. Ignored in favor of $OPENZGY_TOKEN.")
        else:
            print("NOTE: $HACK_SAUTH_TEST_TOKEN and $OPENZGY_TOKEN are both set.")

class SDConfig(Config):

    def __init__(self, context):
        """
        Process an iocontext for seismic store, doing consistency checks
        and applying fallbacks from environment variables and hard coded
        defaults.

        The context itself should be a dict or equivalent. The code also
        supports the older style using a class instance with attributes.
        That feature is deprecated and will be removed. It violates the
        principle of least surprise.

        A fully specified iocontext has the following attributes:

            sdurl: string
                Where to contact the seismic store service.
                Defaults to $OPENZGY_SDURL.

            sdapikey: string
                Authorization for application to access the seismic store API.
                Defaults to $OPENZGY_SDAPIKEY.

            sdtoken: string
                User credentials. Set to $OPENZGY_TOKEN if not found,
                beware that this might not be secure. The code will no longer
                use the token last saved by sdcfg as a fallback. If this is
                desired you must specify "FILE:carbon.slbapp.com" as the token.
                Caveat: The sdcfg token is not refreshed so it might time out
                after an hour. Run "sdutil auth idtoken > /dev/null" to refresh.

            maxsize: int specified in MB between 0 and 1024.
                Zero is taken to mean do not consolidate.
                Tell the reader to try to consolidate neighboring bricks
                when reading from seismic store. This is usually possible
                when the application requests full traces or at least traces
                traces longer then 64 samples. Setting maxsize limits this
                consolidation to the specified size. The assumption is that
                for really large blocks the per-block overhead becomes
                insignificant compared to the transfer time.

                Consolidating requests has higher priority than using
                multiple threads. So, capping maxsize might allow more
                data to be read in parallel.

                Note that currently the spitting isn't really smart. With a
                64 MB limit and 65 contiguous 1 MB buffers it might end up
                reading 64+1 MB instead of e.g. 32+33 MB.

                Note that the low level reader should not assume that
                requests are capped at this size. They might be larger
                e.g. when reading the header information.

                Defaults to $OPENZGY_MAXSIZE_MB if not specified, or 2 MB.

            maxhole: int specified in MB between 0 and 1024.
                This applies when consolidate neighboring bricks when
                reading from seismic store. Setting maxhole > 0 tells the
                reader that it is ok to also consolidate requests that are
                almost neighbors, with a gap up to and including maxhole.
                The data read from the gap will be discarded unless picked
                up by some (not yet implemented) cache.

                For cloud access with high bandwidth (cloud-to-cloud) this
                should be at least 2 MB because smaller blocks will take
                just as long to read. For low bandwidth cloud access
                (cloud-to-on-prem) it should be less. If a fancy cache
                is implemented it should be more. For accessing on-prem
                ZGY files it probably makes no difference.
                Defaults to $OPENZGY_MAXHOLE_MB if not specified, or 2 MB.

            aligned: int in MB between 0 and 1024.
                This is similar to the maxhole parameter. If set, starting
                and ending offsets are extended so they both align to the
                specified value. Set this parameter if the lower levels
                implement a cache with a fixed blocksize and when there is
                an assumpton that most reads will be aligned anyway.
                TODO-Worry: Handling reads past EOF may become a challenge
                for the implementation.
                Defaults to $OPENZGY_ALIGNED_MB if not specified, or zero.

            segsize: int in MB between 0 and 16*1024 (i.e. 16 GB).
                Defaults to $OPENZGY_SEGSIZE_MB if not specified, or 1 GB.

            threads: int between 1 and 1024.
                Use up to this many parallel requests to seismic store
                in order to speed up processing. This applies to individual
                reads in the main API. So the reads must be for a large
                area (i.e. covering many bricks) for the setting to be
                of any use. Set to $OPENZGY_NUMTHREADS if not found,
                and 1 (i.e. no threading) if the environment setting is
                also missing.

                Whether it is useful to set the variable depende on the
                application. Apps such as Petrel/BASE generally do their
                own multi threading, issuing multiple read requests to
                the high level API in parallel. In that case it might
                not be useful to also parallelize individual requests.

            legaltag: string, possibly empty.
                The legaltag stored in the file. Used only on create.

            writeid:
                I don't know what this is for. Ask the seismic store team.

            seismicmeta:
                a dictionary of additional information to be associated
                with this dataset in the data ecosystem. Currently used
                only on create, although SDAPI allows this to be set on
                an existing file by calling {get,set}SeismicMeta().

                When set via an environment variable (strictly for testing)
                this needs to be the string representation of the json data.
                When set from a program a Python dict is expected.

            _debug_trace:
                For debugging and unit tests only.
                Callback to be invoked immediately before a read or write
                is passed on to seismic store. Typically used to verify
                that consolidating bricks works as expected. Can only be
                set programmatically. Not by an environment variable.
"""
        _checkTokenVariableName()
        self.sdurl    = self._get_string_env(context, "sdurl",
                            "OPENZGY_SDURL",
                            "")
        self.sdapikey = self._get_string_env(context, "sdapikey",
                            "OPENZGY_SDAPIKEY",
                            "")
        self.sdtoken  = self._get_string_env(context, "sdtoken",
                            "OPENZGY_TOKEN",
                            "")
        self.maxsize = self._get_numeric_env(context, "maxsize",
                            "OPENZGY_MAXSIZE_MB",64,0,1024) * (1024*1024)
        self.maxhole = self._get_numeric_env(context, "maxhole",
                            "OPENZGY_MAXHOLE_MB",2,0,1024) * (1024*1024)
        self.aligned = self._get_numeric_env(context, "aligned",
                            "OPENZGY_ALIGNED_MB",0,0,1024) * (1024*1024)
        self.segsize = self._get_numeric_env(context, "segsize",
                            "OPENZGY_SEGSIZE_MB",1024,0,1024*16) * (1024*1024)
        self.threads = self._get_numeric_env(context, "threads",
                            "OPENZGY_NUMTHREADS",1,1,1024)
        # All the numeric options ought to be integral, but for unit
        # tests it might be desirable to allow odd sizes. When reading
        # from environment variables only integral numbers are accepted.
        self.maxsize = int(round(self.maxsize))
        self.maxhole = int(round(self.maxhole))
        self.aligned = int(round(self.aligned))
        self.segsize = int(round(self.segsize))

        self.legaltag = self._get_string_env(
            context, "legaltag", "OPENZGY_LEGALTAG", "")
        self.writeid = self._get_string_env(
            context, "writeid", "OPENZGY_WRITEID", "")
        self.seismicmeta = self._get_string_env(
            context, "seismicmeta", "OPENZGY_SEISMICMETA", "")

        try:
            self._debug_trace = context["_debug_trace"]
        except (TypeError, KeyError):
            if hasattr(context, "_debug_trace"):
                self._debug_trace = getattr(context, "_debug_trace")
            else:
                self._debug_trace = lambda *args, **kwargs: False

    @property
    def extra(self):
        """
        Legaltag, writeid, and seismicmeta are usually packed
        into a single "extra" dictionary when creating a new file.
        If any of them are unset they will be excluded from the
        dictionary instead of being passed as some default value.

        CAVEAT: The keys in the "extra" dictionary are not
        supposed to be hard coded as I do here. They are defined in
        seismic-store-client-api-cpp/src/src/core/Constants.{cc,h}.
        Cannot access that file here.

        NOTE: Python dicts have an undefined sort order, as does
        json. To simplify testing I sort the keys in the "extra" dict.
        If SDAPI for some reason should require a specific ordering
        then "seismicmeta" needs to be passed as a string.
        """
        result = {}
        if self.legaltag:
            result["legal-tag"] = self.legaltag
        if self.writeid:
            result["writeid"] = self.writeid
        if self.seismicmeta:
            if not isinstance(self.seismicmeta, str):
                result["seismicmeta"] = json.dumps(
                    self.seismicmeta, sort_keys=True)
            else:
                result["seismicmeta"] = self.seismicmeta
        return result

    @staticmethod
    def _redact(s):
        return s if len(s) < 20 else s[:4] + "..." + s[-4:]

    def dump(self):
        return """SDConfig:
   sdurl:    {sdurl}
   sdapikey: {sdapikey}
   sdtoken:  {sdtoken}
   maxsize:  {maxsize} MB
   maxhole:  {maxhole} MB
   aligned:  {aligned} MB
   segsize:  {segsize} MB
   threads:  {threads}
   extra:    {extra}""".format(
       sdurl=self.sdurl,
       sdapikey=self._redact(self.sdapikey),
       sdtoken=self._redact(self.sdtoken),
       maxsize=self.maxsize // (1024*1024),
       maxhole=self.maxhole // (1024*1024),
       aligned=self.aligned // (1024*1024),
       segsize=self.segsize // (1024*1024),
       threads=self.threads,
       extra=json.dumps(self.extra, sort_keys=True))

class FileADT:

    def __init__(self, filename, mode, iocontext):
        """
        Open a file in the specified mode, which must be "rb" or "w+b".
        Caller should use a "with" block to ensure the file gets closed.
        The iocontext is an optional data structure that the user may
        specify when a reader is created. It might be used to hold
        user credentials etc. needed to access the low level file.
        TODO-Low: support "r+b" (update) at some point in the future.
        """
        if not mode in ("rb", "w+b", "r+b", "d"):
            raise ZgyUserError("Opening ZGY as " + mode + " is not supported.")
        # Currently no need to keep this, as the derived classes will
        # copy the relevant information to self._config.
        #self._iocontext = iocontext

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.xx_close()

    def xx_close(self):
        """
        Close a previously opened file.
        No action if the file is already closed.
        """
        raise NotImplementedError

    def xx_read(self, offset, size, *, usagehint=UsageHint.Unknown):
        """
        Read binary data from the file. Both size and offset are mandatory.
        I.e. caller is not allowed to read "the entire file", and not
        allowed to "read from where I left off the last time".
        The actual reading will be done in a derived class.
        The base class only validates the arguments.
        """
        raise NotImplementedError

    def xx_write(self, data, offset, *, usagehint=UsageHint.Unknown):
        """
        Write binary data to the file. Offset is mandatory. I.e. caller
        is not allowed to "write to where I left off the last time".
        The actual writing will be done in a derived class.
        The base class only validates the arguments.
        """
        raise NotImplementedError

    # Might want to pass this tuple to the delivery functor instead of
    # discrete arguments. Probably doesn't change much either way, though.
    # Or maybe consider passing everything as keyword arguments.
    _deliveryType = namedtuple("Delivery", "data this_offset this_size want_offset want_size iocontext")
    _requestType = namedtuple("Request", "offset size delivery")

    def xx_readv(self, requests, *, parallel_ok=False, immutable_ok=False, transient_ok=False, usagehint=UsageHint.Unknown):
        """
        Read binary data from multiple regions in the file. Each part
        of the request specifies offset, size, and a delivery functor
        which will be invoked to pass back the returned bulk.

        Arguments:
            parallel_ok:  If true then the delivery functor might be called
                          simultaneously from multiple worker threads.
                          The function itself will block until all the data
                          has been read or an error occurs.
            immutable_ok: If true the caller promises that the delivery
                          functor will not try to modify the data buffer.
                          Pass False e.g. if the functor may need to byteswap
                          the data it has read from file.
            transient_ok: If true the caller promises that the delivery
                          functor will not keep a reference to the data buffer
                          after the functor returns.

        The delivery functor is called as
            fn(data)

        FUTURE: a new argument partial_ok may be set to True if it is ok to
        call the delivery functor with less data than requested, and to keep
        calling it until all data has been delivered. The signature of the
        delivery functor gets changed to fn(data, offset, size). Offset is the
        absolute file offset. I.e. not relative to the requested offset.
        Passing partial_ok=True might elide some buffer copies if the
        caller is doing something simple (such as reading an uncompressed
        brick) where partial copies are possible, and the backend is in the
        cloud, and a longer lived cache is being maintained, and the cache
        block size is smaller than the requested size. That is a lot of ifs.
        There was some code to handle partial_ok but it has been removed.
        Get it from the git history if you really want it.
        """
        raise NotImplementedError

    @staticmethod
    def _nice(n):
        """Human readable number."""
        if n >= 1024*1024 and (n % (1024*1024)) == 0:
            return str(n//(1024*1024)) + " MB" # whole number of MB
        elif n >= 256*1024 and (n % (256*1024)) == 0:
            return str(n/(1024*1024)) + " MB" # e.g. 42.75 NB
        elif n >= 1024 and (n % 1024) == 0:
            return str(n/1024) + " kB"
        else:
            return str(n) + " bytes"

    def _validate_read(self, offset, size):
        if self._mode not in ("rb", "w+b", "r+b"):
            raise ZgyUserError("The file is not open for reading.")
        if offset is None or offset < 0:
            raise ZgyUserError("Invalid offset {0} {1} for reading.".format(
                str(type(offset)), str(offset)))
        if size is None or size < 1:
            raise ZgyUserError("Invalid size {0} {1} for reading.".format(
                str(type(size)), str(size)))
        # Beware of mixing python and numpy scalars on Windows.
        # If offset fits in np.int32 then this is what it gets
        # cast to, which could make the sum overflow. On Linux
        # with a slightly older Python but same numpy version
        # the sum always ends up as np.int64. However, in this
        # particular case the exception should never occur so
        # the consequence was only less robust code.
        if int(offset) + int(size) > self.xx_eof:
            # Might be an internal error or a corrupted file,
            # but let's report only the immediate error and not
            # try to guess.
            raise ZgyEndOfFile("Offset {0} size {1} is past EOF at {2}".format(
                self._nice(offset), size, self._nice(self.xx_eof)))

    def _validate_write(self, data, offset):
        if self._mode not in ("w+b", "r+b"):
            raise ZgyUserError("The file is not open for writing.")
        if offset is None or offset < 0:
            raise ZgyUserError("Invalid offset for writing.")
        if data is None or len(data) < 1:
            raise ZgyUserError("Invalid size for writing.")

    def _validate_readv(self, requests):
        for offset, size, delivery in requests:
            self._validate_read(offset, size)

    def _check_short_read(self, offset, size, got):
        """
        Throw a descriptive error if there was something wrong with the read.
        Currently works for local files only.
        """
        # TODO-Low, can I get this to work also for seismic store?
        # Either make a virtual _get_real_size(), or admit this is
        # local-file spacific and move it down to class LocalFile.

        # Work around issue with mixing numpy and Python ints.
        offset, size, got = (int(offset), int(size), int(got))
        if got == size:
            return
        msg = "Cannot read offset {0} size {1}: ".format(
            self._nice(offset), self._nice(size))
        if got > size:
            raise ZgyInternalError(msg + "got too much data: {0}.".format(
                self._nice(got)))
        elif offset + size > self.xx_eof:
            # This can only happen if I (bug!) forgot to call _validate_read.
            raise ZgyEndOfFile(msg + "past EOF at {2}.".format(
                self._nice(self.xx_eof)))
        elif os.stat(self._file.fileno()).st_size < self.xx_eof:
            # This can happen if opening /dev/null for read/write,
            # or if a write failed due to a full disk (and was not checked),
            # or I somehow (bug!) failed to keep track of eof while writing.
            # Or maybe some other process truncated the file.
            raise ZgyEndOfFile(msg + "File is shorter than expected: {0} / {1}.".format(
                self._nice(os.stat(self._file.fileno()).st_size),
                self._nice(self.xx_eof)))
        else:
            # The os returned a short read for no apparent reason.
            # Maybe the file is a special device other than /dev/null.
            raise ZgyEndOfFile(msg + "short read for unknown reason.")

    @staticmethod
    def _consolidate_requests(requests, *,
                              max_hole = 2*1024*1024,
                              max_size = 64*1024*1024,
                              force_align = None,
                              consolidate_overlaps = False,
                              eof = None):
        """
        Given a list of requests as passed to xx_readv, try to reduce
        the number of requests by consolidating adjacent or nearly
        adjacent reads. If successful this means we will be reading
        with larger block sizes.

        Return a new list of requests that the caller may pass on
        to xx_readv instead of the original.

        Remember that the callback functions specified with the
        original requests need to be called with the exact data
        that they expected. This means that in the consolidated
        list the callback functions need to be wrappers.

        Parameters:
            max_hole:    See class SDConfig for a description.
            max_size:    See class SDConfig for a description.
            force_align: See class SDConfig for a description.
            consolidate_overlaps: Set to True if you expect some of the
                      individual requests to overlap, and you are ok with
                      risking some corner cases. For example, if you
                      request a mutable buffer then the overlapping area
                      will be delivered to more than one recipient and
                      the buffer may or may not be shared between the two.
                      The default is False which causes the code to not
                      attempt consolidation of these. Less efficient
                      but also less surprises. In practice there should
                      never be any overlap anyway.
        """
        class ConsolidatedDelivery:
            """
            Helper function to distribute a single delivery from
            a consolidated read request to all the requesters.
            Slice the data so each requester gets exactly what
            they originally asked for.
            Note that if the delivered data is bytes or bytearray
            the slicing will create a copy. If it is a numpy array
            the slice will just return a more efficient view.
            Should I perhaps create a numpy array here?
            Note that if the original request had overlapping reads
            we might want to force a copy anyway. Because we don't
            know whether the recipient asked for a mutable buffer.
            It is tempting to disallow overlapping reads completely.
            Caveat: Handling reads past EOF may be tricky.
            I need some specific unit tests for that.
            """
            def __init__(self, group, begin):
                self._group = group
                self._begin = begin
            def __call__(self, data):
                for offset, size, delivery in self._group:
                    if delivery:
                        end = min(offset + size - self._begin, len(data))
                        beg = min(offset - self._begin, end)
                        delivery(data[beg:end])

        def _groupsize(g, force_align, *, eof):
            """
            Given a list of (offset, size, functor)
            return offset and size for the entire group.
            The offset is the linear offset from the start of the file;
            it has not yet been converted to segment and local offset.
            The returned value includes any padding for force_align.

            TODO-High the padding is WRONG, because the alignment should be
            done per segment. We may end up crossing segment boundaries
            needlessly. And/or going past EOF. Going past EOF is critical
            because in the subsequent call to _split_by_segment() we will
            end up trying to actually read that part.

            Crossing segment boundaries is less of a problem.

              - It will not happen if the headers are aligned at least to
                force_align, which is typically the cache bricksize.

              - It will not happen if the file was uploaded with sdutil.
                In that case there will be just one segment.

              - It is (alomst) not an issue if a proper cache is attached.

              - A naive cache can align to 256 KB, this virtually guarantees
                the header area will be sufficiently aligned if the file
                was initially stored on the cloud.

              - In other cases there will be a small performance penalty but
                only when reading close to a segment boundary or when reading
                the headers. Opening a file may see a noticeable slowdown
                but not I think anything dramatic.
            """
            beg = min([e[0] for e in g])
            end = max([e[0] + e[1] for e in g])
            assert beg == g[0][0]
            #assert end == g[-1][0] + g[-1][1] # might fail if overlap.
            if force_align:
                beg = (beg // force_align) * force_align
                end = ((end + force_align - 1) // force_align) * force_align
                if eof: end = min(end, eof)
            return beg, end - beg

        def _split_requests(requests, force_align, *, eof):
            """
            Make a list of lists, grouping requests that should be read
            in a single operation. Operates on linear addresses, so if
            any of the input requests crossed a segment boundary then
            this will also be the case for the output.
            """
            all_requests = []
            prev_request = (0, 0, None)
            for request in sorted(requests):
                hole = request[0] - (prev_request[0] + prev_request[1])
                if not all_requests:
                    all_requests = [[request]]
                elif (hole <= max_hole and
                      (consolidate_overlaps or hole >= 0) and
                      (not max_size or _groupsize(all_requests[-1] + [request], force_align, eof=eof)[1] <= max_size)):
                    all_requests[-1].append(request)
                else:
                    all_requests.append([request])
                prev_request = request
            return all_requests

        def _join_requests(all_requests, force_align, *, eof):
            """Create the final result containing one entry for each
            consolidated group."""
            new_requests = []
            for group in all_requests:
                # Short cut, probably not worth the trouble.
                #if len(group)==1 and not force_align: new_requests.append(group[0])
                offset, size = _groupsize(group, force_align, eof=eof)
                new_requests.append((offset, size, ConsolidatedDelivery(group, offset)))
            return new_requests

        def _print_requests(all_requests, name = "Requests:"):
            """For debugging only, print a list of list of requests."""
            if len(all_requests) == 0 or (len(all_requests) == 1 and len(all_requests[0]) == 0):
                print("    (empty)")
                return
            print(name)
            for group in all_requests:
                if len(all_requests) > 1: print("  Group:")
                prev_offset, prev_size = (None, None)
                msg = "    {0} offset {1:8X} end {2:8X} size {3:6X}"
                for offset, size, delivery in group:
                    if prev_offset is not None:
                        skip_offset = prev_offset + prev_size
                        skip_size = offset - (prev_offset + prev_size)
                        if skip_size != 0:
                            print(msg.format("skip", skip_offset, offset, skip_size))
                    print(msg.format("read", offset, offset+size, size))
                    prev_offset, prev_size = (offset, size)


        # main part of _consolidate_requests().
        all_requests = _split_requests(requests, force_align, eof=eof)
        new_requests = _join_requests(all_requests, force_align, eof=eof)
        if False and len(requests) != len(new_requests):
            print("Consolidated", len(requests), "into", len(new_requests))
            print("Requests:")
            _print_requests([requests])
            print("Consolidated:")
            _print_requests([new_requests], name="Consolidated:")
        old_size = sum([x[1] for x in requests])
        new_size = sum([x[1] for x in new_requests])
        assert new_size >= old_size
        return new_requests

    @property
    def threadsafe(self):
        return False

    @property
    def xx_iscloud(self):
        return False

class LocalFile(FileADT):
    def __init__(self, filename, mode, iocontext):
        super().__init__(filename, mode, iocontext)
        self._config = FileConfig(iocontext)
        self._file = open(filename, mode) if mode != "d" else None
        self._mode = mode
        self._name = filename
        self._eof = 0 if mode in ("wb", "w+b") else os.stat(filename).st_size

    @property
    def xx_eof(self):
        return self._eof

    def xx_close(self):
        mode = self._mode
        self._mode = None
        if mode and self._name:
            if mode == "d":
                with suppress(FileNotFoundError):
                    os.remove(self._name)
            else:
                self._file.close()
        self._file = None
        self._name = None

class LocalFileOther(LocalFile):
    def xx_read(self, offset, size, *, usagehint=UsageHint.Unknown):
        self._validate_read(offset, size)
        self._file.seek(offset, 2 if offset < 0 else 0)
        data = self._file.read(size)
        self._check_short_read(offset, size, len(data))
        return data

    def xx_readv(self, requests, *, parallel_ok=False, immutable_ok=False, transient_ok=False, usagehint=UsageHint.Unknown):
        for offset, size, delivery in requests:
            delivery(LocalFileOther.xx_read(self, offset, size, usagehint=usagehint))

    def xx_write(self, data, offset, *, usagehint=UsageHint.Unknown):
        self._validate_write(data, offset)
        self._file.seek(offset)
        self._eof = max(self._eof, offset + len(data))
        nbytes = self._file.write(data)
        if nbytes != len(data):
            raise ZgyInternalError("Short write to local file")
        return len(data) # Most callers ignore this anyway.

    @property
    def threadsafe(self):
        return False

class LocalFileLinux(LocalFile):
    def xx_read(self, offset, size, *, usagehint=UsageHint.Unknown):
        self._validate_read(offset, size)
        data = os.pread(self._file.fileno(), size, offset)
        self._check_short_read(offset, size, len(data))
        return data

    def xx_readv(self, requests, *, parallel_ok=False, immutable_ok=False, transient_ok=False, usagehint=UsageHint.Unknown):
        for offset, size, delivery in requests:
            delivery(LocalFileLinux.xx_read(self, offset, size, usagehint=usagehint))

    def xx_write(self, data, offset, *, usagehint=UsageHint.Unknown):
        self._validate_write(data, offset)
        self._eof = max(self._eof, offset + len(data))
        nbytes = os.pwrite(self._file.fileno(), data, offset)
        if nbytes != len(data):
            raise ZgyInternalError("Short write to local file")

    @property
    def threadsafe(self):
        return True

class SeismicStoreFile(FileADT):
    """
    Access data in seismic store as a linear file even when the dataset
    has multiple segments. There are some limitations on write.

      * Writes starting at EOF are allowed, and will cause a new segment
        to be written.

      * Writes starting past EOF, signifying a hole in the data, are not
       allowed.

      * Writes starting before EOF are only allowed if offset,size exactly
        matches a previous write. This will cause that segment to be rewritten.

      * Possible future extension: For the  last segment only offset
        needs to match. This means the last segment may be resized.

    For read the class provides a readv() method to do scatter/gather reads.
    The code will then consolidate adjacent bricks to get larger brick size
    sent to SDAPI. Optionally parallelize requests that cannot be consolidated.
    """
    def __init__(self, filename, mode, iocontext):
        if sd is None:
            raise ZgyMissingFeature("Seismic Store is not available")
        super().__init__(filename, mode, iocontext)
        self._config = SDConfig(iocontext)
        #print(self._config.dump())

        sdcred = (self._config.sdurl, self._config.sdapikey, self._config.sdtoken)
        if not all(sdcred):
            raise ZgyUserError("Missing credentials:" +
                               ("" if sdcred[0] else " $OPENZGY_SDURL") +
                               ("" if sdcred[1] else " $OPENZGY_SDAPIKEY") +
                               ("" if sdcred[2] else " $OPENZGY_TOKEN"))
        if mode in ("rb"):
            self._accessor = sd.SdReader(filename, sdcred)
            # Get the size of each segment. For efficiency assume that all
            # segments except the first and last will have the same size.
            # TODO-Medium: If sizes() is cheap just request all of them.
            numseg = self._accessor.count()
            if numseg <= 3:
                self._sizes = list(self._accessor.sizes(*range(numseg)))
            else:
                tmp = self._accessor.sizes(0, 1, numseg-1)
                self._sizes = [tmp[0]] + (numseg-2) * [tmp[1]] + [tmp[2]]
        elif mode in ("w+b"):
            #print(self._config.dump(), sep="\n")
            # Create new, deleting or truncating existing.
            self._accessor = sd.SdWriter(filename, sdcred, False, self._config.extra)
            # TODO-Medium: If the file existed already, the mutable parts of the
            # metadata is allowed to change. Data can only be written if
            # the file was completely empty i.e. with just segment 0.
            self._sizes = []
        elif False and mode in ("r+b"):
            # TODO-High open without truncating not supported yet.
            # Limited support would be to open a file with only headers.
            # Full support is much trickier, need to re-open last segment
            # and also preferably do an incremental update of statistics
            # and lowres bricks.
            # Also, legaltag (and maybe seismicmeta) might be ignored here.
            self._accessor = sd.SdWriter(filename, sdcred, True, self._config.extra)
            numseg = self._accessor.count()
            if numseg <= 3:
                self._sizes = list(self._accessor.sizes(*range(numseg)))
            else:
                tmp = self._accessor.sizes(0, 1, numseg-1)
                self._sizes = [tmp[0]] + (numseg-2) * [tmp[1]] + [tmp[2]]
        elif mode in ("d"):
            # TODO-Performance keep the SdUtil instance alive.
            # There is a cost creating a new one, especially if the
            # token needs a refresh.
            self._accessor = None
            try:
                with sd.SdUtil(sdcred) as u:
                    u.delete(filename)
            except RuntimeError as ex:
                # File not found is ignored. Other errors are reported back.
                mode = None
                if str(ex).find("does not exist") < 0:
                    raise
        else:
            raise ZgyUserError("Opening ZGY as " + mode + " is not supported.")
        self._mode = mode
        self._cached_data = None

    @property
    def xx_eof(self):
        return sum(self._sizes)

    def xx_close(self):
        mode = self._mode
        self._mode = None
        if self._accessor and mode and mode != "d":
            self._accessor.close()
        self._accessor = None

    def _split_by_segment(self, requests):
        """
        Given one or more (offset, size, ...) tuples, convert these
        to (segment_number, offset_in_seg, size_in_seg, outpos).

        "outpos" is the offset to store the data that was read, if
        it is to be stored sequentially in one large buffer.

        Request for data past EOF is converted to a request for data
        in the last segment plus one. Trying to read that segment from
        seismic store will fail. Note that if _sizes is changed to
        include the open segment at the end then this special handling
        makes not much difference. At least not if the argument has
        already been checked to not cross the real EOF.

        Note that the callers currently check for reads past EOF
        and raises an exception in that case. So for now the above
        paragraph if of academic interest only.

        The returned list might be longer than the input if any of the
        input requests crossed segment boundaries.
        The return might be shorter than the input or even empty if
        any input request was for 0 bytes [or data past EOF... NOT]

        The algorithm is O(n^2) on segment_count * request_count
        but both numbers should be small. If this actually becomes
        a problem then use binary search in self._cumsize to find
        the starting segment.

        Maybe simplify: This logic could be moved inside SDAPI or the
        SDAPI wrapper. Reads from segment "-1" imply linear access.
        There would be a slight change in that requests split due to
        crossing a segment boundary would not be parallelized. But
        that is expected to be a very rare occurrence. Caveat, be
        careful what happens if reading past EOF. The code can currently
        handle that by returning data from the "open segment". That logic
        might not be possible to put into SDAPI. So this is probably
        not a good idea.
        """
        result = []
        outpos = 0
        for entry in requests:
            offset, size = entry[:2]
            assert offset >= 0
            size = entry[1]
            seg = 0
            for segsize in self._sizes:
                this_size = max(0, min(size, segsize - offset))
                if this_size > 0:
                    result.append((seg, offset, this_size, outpos))
                    offset += this_size
                    outpos += this_size
                    size   -= this_size
                    # If not crossing segment boundary, size will be 0
                    # Otherwise offset will be segsize, meaning that in
                    # the next iteration offset will be 0 and size will
                    # be the remaining data to be read.
                seg += 1
                offset -= segsize
                if size <= 0: break
        if size > 0:
            # Data past EOF treated as if it were in segment+1.
            result.append((seg, offset, size, outpos))
        insize = sum([e[1] for e in requests])
        outsize = result[-1][2] + result[-1][3] if result else 0
        assert insize == outsize
        return result

    def _cached_read(self, seg, offset, view):
        """
        Very simplistic cache implementation. Only store the most recent
        read from seismic store, and only consider a match when the range
        matches exactly. Also, always make copies both when copying in to
        and out of the cache. I.e. ignore immutable_ok, transient_ok.

        The cache may be useful if the upstream caller only asks for a
        single block at a time, so we get neither consolidation nor
        parallel access. Enabling this cache and setting force_align
        to a suitable value will hopefully cause the code to see the
        same bulk request happen more than once. If force_align is off
        it is very unlikely that the cache will help.

        Understand that storing just a single block in the cache will
        only help in lab conditions or in applications that we know
        for sure will issue requests for sequential parts of the cube.
        And if we know this already then we ought to be able to modify
        that code to pass down larger requests. Bottom line, it isn't
        very useful the way it works today.
        """
        if not self._config.aligned:
            self._config._debug_trace("read", len(view), len(view), 1)
            return self._accessor.read(seg, offset, view)
        seg_size = self._sizes[seg]
        a_beg = (offset // self._config.aligned) * self._config.aligned
        a_end = ((offset + len(view) + self._config.aligned - 1) // self._config.aligned) * self._config.aligned
        a_end = min(a_end, seg_size)
        c_seg, c_beg, c_data = self._cached_data or (0, 0, bytes())
        need = (seg, a_beg, a_end)
        have = (c_seg, c_beg, c_beg + len(c_data))
        #print('cache', need, ('==' if need == have else '<>'), have)
        if need == have:
            self._config._debug_trace("cachehit", len(view), a_end-a_beg, 1)
            data = c_data
        else:
            self._config._debug_trace("cachemiss", len(view), a_end-a_beg, 1)
            data = bytearray(a_end - a_beg)
            self._accessor.read(seg, a_beg, data)
            self._cached_data = (seg, a_beg, data)
        view[:] = data[offset-a_beg:offset-a_beg+len(view)]

    def xx_read(self, in_offset, in_size, *, usagehint=UsageHint.Unknown):
        self._validate_read(in_offset, in_size)
        work     = self._split_by_segment([(in_offset, in_size)])
        result   = bytearray(in_size)
        view     = memoryview(result)
        maxseg = max([seg for seg, offset, size, outpos in work])
        if maxseg >= len(self._sizes):
            # This should only happen in white box unit tests.
            # The higher levels of ZGY should have checked for EOF already.
            # But seismic store can return a really obscure error message
            # and/or hang in a retry loop if we don't do this check.
            raise ZgyEndOfFile("Attempt to read from segment " + str(maxseg))
        for seg, offset, size, outpos in work:
            self._cached_read(seg, offset, view[outpos:outpos+size])
        return result

    def xx_readv(self, requests, *, parallel_ok=False, immutable_ok=False, transient_ok=False, usagehint=UsageHint.Unknown):
        """
        Handle both brick consolidation and multi threading.

        This implementation will issue a single readv() request to the
        seismic store wrapper, wait for all threads to complete, and
        then deliver all the results. For this reason it needs to
        allocate a buffer to hold the entire data to be read.

        In the future it might be possible to have the seismic store
        wrapper support delivery callbacks and for it to allocate
        the result buffers itself. This saves some memory and also
        allows data to be decompressed if needed and copied out to
        user space as the bricks become available. Caveat: requests
        may need to be split if they cross a segment boundary.
        This means that we need support for partial delivery.
        Which would complicate things a lot.
        """
        self._validate_readv(requests)

        # I don't really like this kind of short cut since it creates
        # a lot of corner cases to test for. But, if the naive caching
        # is in effect then it is needed to make caching work.
        # If multiple requests then the cache won't work anyway,
        # and we might as well clear any data it contains.
        # TODO-Performance, can I move this test after consolidate
        # and split? Otherwise it will probably only work for the headers
        # and when the application really did fetch just one brick at
        # time. It might be good enough for Petrel though.
        if self._config.aligned and len(requests) == 1:
            for offset, size, delivery in requests:
                delivery(SeismicStoreFile.xx_read(self, offset, size, usagehint=usagehint))
            return
        self._cached_data = None

        # For debugging / logging only
        asked = sum([e[1] for e in requests])

        new_requests = self._consolidate_requests(requests,
                                                  max_hole=self._config.maxhole,
                                                  max_size=self._config.maxsize,
                                                  force_align=self._config.aligned,
                                                  eof=self.xx_eof)
        work     = self._split_by_segment(new_requests)
        realsize = work[-1][2] + work[-1][3] if work else 0
        data     = bytearray(realsize)
        view     = memoryview(data)
        eof      = sum(self._sizes)

        # Read bulk data from seismic store using multiple threads.
        self._config._debug_trace("readv", asked, len(view), len(work))
        self._accessor.readv(work, data, self._config.threads)

        # Deliver result to caller.
        pos = 0
        for offset, size, delivery in new_requests:
            size = max(0, min(size, eof - offset))
            delivery(view[pos:pos+size])
            pos += size

    def xx_write(self, data, offset, *, usagehint=UsageHint.Unknown):
        self._validate_write(data, offset)
        current_eof = SeismicStoreFile.xx_eof.__get__(self) # nonvirtual call
        #print("SeismicStoreFile.xx_write(offset={0}, size={1}, current EOF is {2}".format(offset, len(data), current_eof))
        if offset == current_eof:
            # Sequential write from current EOF.
            # Specific limitation for ZGY, for performance reasons only.
            # This means we don't need to fetch sizes for all segments
            # when opening a file for read. Note that since the last
            # segment can have any size we won't discover a violation
            # until the following read.
            if len(self._sizes) >= 3 and self._sizes[-1] != self._sizes[1]:
                raise ZgyUserError("Cannot write arbitrarily sized segment.")
            self._config._debug_trace("append", len(data), len(data), 1)
            self._accessor.write(len(self._sizes), data, False)
            self._sizes.append(len(data))
        elif offset < current_eof:
            # Rewrite existing block. Resizing not allowed.
            seg = 0
            for segsize in self._sizes:
                if offset == 0:
                    if len(data) == segsize:
                        self._config._debug_trace("write", len(data), len(data), 1)
                        self._accessor.write(seg, data, True)
                        break
                    else:
                        raise ZgySegmentIsClosed("Cannot write resized segment.")
                elif offset < segsize:
                    raise ZgySegmentIsClosed("Cannot write part of segment.")
                seg += 1
                offset -= segsize
        else:
            # Attempting to write sparse data.
            raise ZgyUserError("Cannot write segments out of order.")
        return len(data)

    # If I want to disable threading, possibly also consolidation:
    #xx_readv = FileADT._forward_consolidated_readv
    #xx_readv = FileADT._forward_readv

    @property
    def threadsafe(self):
        return True if self._mode in ("rb") else False

    @property
    def xx_iscloud(self):
        return True

class SeismicStoreFileDelayedWrite(FileADT):
    """
    Improve on SeismicStoreFile, have it buffer large chunks of data before
    writing it out to a new segment.

      * Writes starting at EOF are allowed, and will buffer data in the
       "open segment" until explicitly flushed.

      * Writes starting past EOF, signifying a hole in the data, are not
       allowed.

      * Writes fully inside the open segment are allowed.

      * Writes starting before the open segment are only allowed if
        offset,size exactly matches a previous write. This will cause that
        segment to be rewritten. As a corollary, writes canot span the
        closed segment / open segment boundary.

      * Possible future extension: For the  last segment only offset
        needs to match. This means the last segment may be resized.
        Why we might want this: On opening a file with existing
        data bricks we might choose to read the last segment and
        turn it into an open segment. Then delete (in memory only)
        the last segment. When it is time to flush the data it gets
        rewritten. This allows adding bricks to a file, while still
        ensuring that all segments except first and last need to be
        the same size. Note that there are other tasks such as
        incrementally updating statistics and histogram that might
        turn out to be a lot of work.

      * When used to create ZGY files, caller must honor the convention
        that all segments except the first and last must have the same size.

      * Caveat: The fact that random writes are sometimes allowed, sometimes
        not depending on the segment number violates the principle of
        least surprise. And makes for more elaborate testing. For ZGY
        it is quite useful though. ZGY can recover from a ZgySegmentIsClosed
        exception by abandoning (leaking) the current block and write it
        to a new location. With a typical access pattern this will happen
        only occasionally.
    """
    def __init__(self, filename, mode, iocontext):
        super().__init__(filename, mode, iocontext)
        self._relay = SeismicStoreFile(filename, mode, iocontext)
        self._mode = mode
        self._open_segment = bytearray()
        self._usage_hint = None
        self._config = self._relay._config

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.xx_close()

    def _flush_part(self, this_segsize):
        """
        Flush "this_segsize" pending bytes. Leave any residual data
        in the open segment buffer.
        """
        assert this_segsize <= len(self._open_segment)
        assert len(self._open_segment) > 0
        flushme = memoryview(self._open_segment)
        nbytes = self._relay.xx_write(flushme[:this_segsize],
                                      self._relay.xx_eof,
                                      usagehint=self._usage_hint)
        if nbytes != this_segsize:
            raise ZgyInternalError("Short write to seismic store")
        self._open_segment = bytearray(flushme[this_segsize:])

    def _flush(self, final):
        """
        Flush pending writes, but only if we have enough data to fill
        one or more complete segments or if the file is being closed.
        The last segment is allowed to be smaller than the others.
        """
        if self._config.segsize > 0:
            while len(self._open_segment) >= self._config.segsize:
                self._flush_part(self._config.segsize)
        if final and len(self._open_segment) > 0:
            self._flush_part(len(self._open_segment))
        if len(self._open_segment) == 0:
            self._usage_hint = None

    @property
    def xx_eof(self):
        """
        Current size of the zgy file, including any buffered unwritten data.
        """
        return self._relay.xx_eof + len(self._open_segment)

    def xx_write(self, data, offset, *, usagehint=UsageHint.Unknown):
        """
        Write data to seismic store, buffering the writes to get larger
        segment sizes. Writes are only allowed at offset 0 and at EOF.
        This is less general then the parent type which lets us rewrite
        any segment as long as its size does not change.

        Segment 0 contains just the headers and is always written in one
        operation, so this is not buffered. Segment 0 can be both smaller
        and larger than segsize. Which is another reason to bypass the
        buffering code. Also, if we are rewriting data we bypass the
        buffering and require that the caller updates the entire segment.
        ZGY will currently only rewrite segment 0.

        If segsize is zero no buffering is done and each write will either
        create a new segment or completely rewrite an existing segment.
        """
        written = self.xx_eof
        committed = self._relay.xx_eof
        #print("SeismicStoreFileDelayedWrite.xx_write(offset={0}, size={1}, current EOF is {2}".format(offset, len(data), self.xx_eof))
        # TODO-Low: Refactor: Technically I could buffer segment 0 as well,
        # and leave it to the caller to immediately flush that segment.
        # I probably still need some special handling for the first segment.
        # The benefit is that FileADT gets less knowledge about ZGY proper.
        if offset == 0 or self._config.segsize <= 0 or offset < committed:
            return self._relay.xx_write(data, offset, usagehint=usagehint)
        if offset > written:
            # Write sparse data with a hole between written and offset.
            raise ZgyUserError("Cannot write segments out of order.")
        # TODO-Low: Generalize: If caller doesn't catch ZgySegmentIsClosed then all
        # rewrites ought to be forbidden. Since ZGY is our only client and
        # does in fact catch that exception then this is low priority.
        #if offset != self.xx_eof:
        #    raise ZgyUserError("Can only write at offset 0 or EOF, not {0} when EOF is {1}. Also, size is {2} and current open segment has {3} bytes.".format(offset, self.xx_eof, len(data), len(self._open_segment)))
        # TODO-Low: Refactor: Python abuse! Passing a "self" that is not an
        # instance of FileADT. But it has the attributes expected by the method.
        FileADT._validate_write(self, data, offset)
        lll = len(self._open_segment)
        if offset == written:
            # Append data to open segment
            self._open_segment += data
        elif offset + len(data) <= written:
            # Update data fully inside open segment
            self._open_segment[offset-committed:offset-committed+len(data)] = bytearray() + data
        else:
            # part update, part new.
            raise NotImplementedError() # TODO-Low support for symmetry.
        if self._usage_hint is None:
            self._usage_hint = usagehint
        elif self._usage_hint != usagehint:
            self._usage_hint = UsageHint.Unknown # mixed hints
        self._flush(False)
        return len(data) # TODO-Low: retval not useful if I throw on short reads.

    def xx_read(self, offset, size, *, usagehint=UsageHint.Unknown):
        FileADT._validate_read(self, offset, size)
        closed_size = max(0, min(size, self._relay.xx_eof - offset))
        opened_size = size - closed_size
        local_offset = max(0, offset - self._relay.xx_eof)
        if local_offset + opened_size > len(self._open_segment):
            raise ZgyUserError("Reading past EOF")
        data1 = self._relay.xx_read(offset, closed_size, usagehint=usagehint) if closed_size > 0 else None
        data2 = memoryview(self._open_segment)[local_offset:local_offset+opened_size] if opened_size > 0 else None
        return data1 + data2 if data1 and data2 else data1 or data2

    def xx_readv(self, requests, *, parallel_ok=False, immutable_ok=False, transient_ok=False, usagehint=UsageHint.Unknown, **kwargs):
        end = max([offset + size for offset, size, delivery in requests])
        if end <= self._relay.xx_eof:
            # The open segment is not involved, so just forward the request.
            self._relay.xx_readv(requests,
                                 parallel_ok=parallel_ok,
                                 immutable_ok=immutable_ok,
                                 transient_ok=transient_ok,
                                 usagehint=usagehint)
        else:
            # Let xx_read handle the requests one at a time.
            # If the requests consisted of both open and closed segments
            # then this is inefficient since SD access won't be paralellized.
            # But that case would be a lot of effort to support and it
            # won't happen often.
            for offset, size, delivery in requests:
                delivery(self.xx_read(offset, size, usagehint=usagehint))

    def xx_close(self):
        self._flush(True)
        return self._relay.xx_close()

    @property
    def threadsafe(self):
        return self._relay.threadsafe

    @property
    def xx_iscloud(self):
        return self._relay.xx_iscloud

def FileFactory(filename, mode, iocontext):
    """
    Return a FileADT instance able to read and/or write to the named file.
    In the future the function might return different types of instances
    e.g. if filename refers to something on the cloud.
    """
    if filename[:5] == "sd://":
        if mode in ("r+b", "w+b"):
            myfile = SeismicStoreFileDelayedWrite(filename, mode, iocontext)
        else:
            myfile = SeismicStoreFile(filename, mode, iocontext)
    elif hasattr(os, "pread"):
        myfile = LocalFileLinux(filename, mode, iocontext)
    else:
        myfile = LocalFileOther(filename, mode, iocontext)
    return myfile

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
Defines exceptions that may be raised by OpenZGY.

These classes are both visible to the OpenZGY public API and referenced
directly from the implementation classes. I apologize for the broken
encapsulation. Re-mapping the exceptions in the API layer didn't seem
worth the trouble.
"""
##@package openzgy.exception
#@brief Exceptions that may be raised by OpenZGY.

##@brief Base class for all exceptions thrown by %OpenZGY.
class ZgyError(Exception):
    """
    Base class for all exceptions thrown by %OpenZGY.
    """
    pass


##@brief Corrupted or unsupported ZGY file.
class ZgyFormatError(ZgyError):
    """
    Corrupted or unsupported ZGY file.

    In some cases a corrupted file might lead to a ZgyInternalError
    or ZgyEndOfFile being thrown instead of this one. Because it isn't
    always easy to figure out the root cause.
    """
    pass

##@brief The ZGY file became corrupted while writing to it.
class ZgyCorruptedFile(ZgyError):
    """
    The ZGY file became corrupted while writing to it.

    No further writes are allowed on this file because a previous write
    raised an exception and we don't know the file's state. Subsequent
    writes will also throw this exception.

    The safe approach is to assume that the error caused the file to
    become corrupted. It is recommended that the application closes and
    deletes the file.
    """
    pass

##@brief Exception that might be caused by the calling application.
class ZgyUserError(ZgyError):
    """
    Exception that might be caused by the calling application.

    Determining whether a problem is the fault of the calling application
    or the %OpenZGY library itself can be guesswork. Application code
    might choose to treat ZgyUserError and ZgyInternalError the same way.
    """
    pass

##@brief Exception that might be caused by a bug in %OpenZGY.
class ZgyInternalError(ZgyError):
    """
    Exception that might be caused by a bug in %OpenZGY.

    Determining whether a problem is the fault of the calling application
    or the %OpenZGY library itself can be guesswork. Application code
    might choose to treat ZgyUserError and ZgyInternalError the same way.

    A corrupt file might also be reported as ZgyInternalError instead of
    the more appropriate ZgyFormatError.
    """
    pass

##@brief Trying to read past EOF.
class ZgyEndOfFile(ZgyError):
    """
    Trying to read past EOF.

    This is always considered an error, and is often due to a corrupted
    ZGY file. So this error should probably be treated as a ZgyFormatError.
    """
    pass

##@brief Exception used internally to request a retry.
class ZgySegmentIsClosed(ZgyError):
    """
    Exception used internally to request a retry.
    A write to the cloud failed because the region that was attempted
    written had already been flushed. And the cloud back-end does not
    allow writing it again. The calling code, still inside the OpenZGY
    library, should be able to catch and recover from this problem.
    """
    pass

##@brief User aborted the operation.
class ZgyAborted(ZgyError):
    """
    User aborted the operation.

    If the user supplied a progress callback and this callback returned
    false then the operation in progress will and by throwing this
    exception. Which means that this is not an error; it is a consequence
    of the abort.
    """
    pass

##@brief Missing feature.
class ZgyMissingFeature(ZgyError):
    """
    Missing feature.

    Raised if some optional plug-in (e.g. some cloud back end or a
    compressor) was loaded or explicitly requested, so we know about
    it, but the plug-in is not operational for some reason.
    """
    pass

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
Measure compression- and decompression speed for data already in memory.
See zfp-speedtest.sh for usage.
Also reports on requested vs. achieved SNR. Does not store the output data.
To visually inspect the compression quality see zfp-comparison.sh.
"""

import numpy as np
import os, sys, math, time, argparse
from zfpy import compress_numpy as new_compress
from zfpy import decompress_numpy as new_decompress
from zgycompress import compress as old_compress_
from zgycompress import inflate as old_decompress_
from ..api import ZgyReader, ZgyWriter, SampleDataType, ProgressWithDots
from ..test.utils import SDCredentials

# TODO-Low consider using iterator.readall. But, is test_zfp still useful?

global_args = None

def old_compress(data, sqnr):
    assert data.shape == (64, 64, 64)
    if global_args.buggy:
        # We will pass our brick indexed il, xl, z which is wrong
        # since the old compressor wants z varying slowest.
        return old_compress_(data, sqnr)
    else:
        # Explicitly make the buffer fortran contiguous.
        # This is probably simpler than reversing the axis.
        data = np.copy(data, order="F")
        return old_compress_(data, sqnr)

def old_decompress(cdata):
    bricksize = (64, 64, 64) # KLUDGE
    tmp = old_decompress_(cdata, bricksize)
    if global_args.buggy:
        return np.frombuffer(tmp, dtype=np.float32).reshape(bricksize)
    else:
        return np.frombuffer(tmp, dtype=np.float32).reshape(bricksize, order="F")

def snr(s, n):
    return 99 if n == 0 else 20 * math.log10(s / n)

def calc_snr(data, ddata):
    signal = np.sum(np.abs(data), dtype=np.float64)
    noise  = np.sum(np.abs(data - ddata), dtype=np.float64)
    return snr(signal, noise)

def calc_snr_from_list(data, ddata):
    signal = 0
    noise = 0
    for ii in range(len(data)):
        signal += np.sum(np.abs(data[ii]), dtype=np.float64)
        noise  += np.sum(np.abs(data[ii] - ddata[ii]), dtype=np.float64)
    return snr(signal, noise)

def zfp_parameters(want_snr, datarange, dtype):
    """
    Express the Petrel style "requested signal/noise ratio"
    in terms of ZFP configuration parameters.
    """
    if datarange == 0 or not np.isfinite(datarange) or want_snr > 70:
        return dict()
    else:
        tolerance = datarange / math.pow(10, (want_snr-8)/20)
        return { "tolerance": tolerance }

def slurp_open_file(r, progress):
    def _roundup(x, step): return ((x + step - 1) // step) * step
    def _rounddown(x, step): return (x // step) * step
    bricksize = tuple(r.bricksize)
    # Since I am testing compression, don't confuse matters
    # by introducing partial bricks.
    blocksize = (bricksize[0],
                 _rounddown(r.size[1], r.bricksize[1]),
                 _rounddown(r.size[2], r.bricksize[2]))
    total = r.size[0] // blocksize[0]
    # No more than 1024 MVoxel to avoid running out of memory (and time).
    # TODO-Low, would be nice to crop both inline and crossline range
    # to have less likelyhood of dead traces.
    print("BS", np.product(blocksize) // (1024*1024), "MB, count", total)
    have_room = 1024*1024*1024 // np.product(blocksize)
    new_total = min(total, max(have_room, 1))
    first_il = ((total - new_total) // 2) * blocksize[0]
    print("Adjust il count from start 0 count {0} to start {1} count {2}".format(total, first_il, new_total))
    total = new_total
    done = 0
    blocks = []
    data = np.zeros(blocksize, dtype=np.float32)
    for ii in range(first_il, total*blocksize[0] + first_il, blocksize[0]):
        r.read((ii, 0, 0), data)
        for jj in range(0, _rounddown(r.size[1], bricksize[1]), bricksize[1]):
            for kk in range(0, _rounddown(r.size[2], bricksize[2]), bricksize[2]):
                blocks.append(np.copy(data[:,jj:jj+bricksize[1],kk:kk+bricksize[2]]))
        done += 1
        if progress: progress(done, total)
    return blocks

def compress_all(blocks, progress):
    cblocks = []
    done = 0
    total = len(blocks)
    for data in blocks:
        if global_args.old:
            cblocks.append(old_compress(data, global_args.sqnr))
        else:
            datarange = np.amax(data) - np.amin(data) if np.all(np.isfinite(data)) else np.nan
            zfp_kwargs = zfp_parameters(global_args.sqnr, datarange, data.dtype)
            cblocks.append(new_compress(data, **zfp_kwargs))
        done += 1
        if progress: progress(done, total)
    return cblocks

def uncompress_all(cblocks, progress):
    result = []
    done = 0
    total = len(cblocks)
    for cdata in cblocks:
        if global_args.old:
            result.append(old_decompress(cdata))
        else:
            result.append(new_decompress(cdata))
        done += 1
        if progress: progress(done, total)
    return result

def slurp(filename):
    starttime = time.time()
    with ZgyReader(filename, iocontext = SDCredentials()) as r:
        blocks = slurp_open_file(r, progress=ProgressWithDots())
        bricksize_in_bytes = int(np.product(r.bricksize) * 4)
    r_time = time.time() - starttime

    for b in blocks:
        assert b.size * b.itemsize == bricksize_in_bytes

    count1 = len(blocks)
    len1 = 4 * np.sum([len(e.flat) for e in blocks], dtype=np.int64)

    starttime = time.time()
    cblocks = compress_all(blocks, progress=ProgressWithDots())
    c_time = time.time() - starttime

    count2 = len(cblocks)
    len2 = np.sum([len(e) for e in cblocks], dtype=np.int64)
    #blocks = None

    starttime = time.time()
    for i in range(global_args.repeat):
        ublocks = uncompress_all(cblocks, progress=ProgressWithDots())
    u_time = (time.time() - starttime) // global_args.repeat

    count3 = len(ublocks)
    len3 = 4 * np.sum([len(e.flat) for e in ublocks], dtype=np.int64)

    assert count1 == count2
    assert count2 == count3
    assert len1 == len3
    for b in ublocks:
        assert 4*len(b.flat) == bricksize_in_bytes
    assert count1 * bricksize_in_bytes == len1

    MB = 1024*1024
    r_rate = (len1 / r_time) / MB if r_time else 0
    c_rate = (len1 / c_time) / MB if c_time else 0
    u_rate = (len1 / u_time) / MB if u_time else 0


    # Now measure the signal to noise ratio:
    snrlist = [calc_snr(blocks[ii], ublocks[ii]) for ii in range(count1)]
    snrlist = list(sorted(snrlist))
    total_snr = calc_snr_from_list(blocks, ublocks)

    info = {
        "name": filename,
        "brickcount": count1,
        "bricksize": bricksize_in_bytes / MB,
        "compressor": ("ZGY" if global_args.old else "ZFP"),
        "r_time": r_time,
        "c_time": c_time,
        "u_time": u_time,
        "r_rate": r_rate,
        "c_rate": c_rate,
        "u_rate": u_rate,
        "u_mb": len1 / MB,
        "c_mb": len2 / MB,
        "c_ratio": len1 / len2,
        "snr_ask": global_args.sqnr,
        "snr_min": snrlist[0],
        "snr_max": snrlist[-1],
        "snr_avg": np.sum(snrlist)/len(snrlist),
        "snr_median": snrlist[len(snrlist)//2],
        "snr_total": total_snr,
    }

    if global_args.csv:
        hdr_format = "!!;Name;Brick count;Brick size (MB);Compressor;Read (sec);Compress (sec);Inflate(sec);Read (MB/s);Compress (MB/s);Inflate(MB/s);Uncompressed (MB);Compressed (MB);Compression Ratio;Requested SNR;Min SNR;Max SNR;Avg SNR;Median SNR;Overall SNR" if global_args.csvheader else None
        data_format = "!!;{name};{brickcount};{bricksize:.0f};{compressor};{r_time:.2f};{c_time:.2f};{u_time:.2f};{r_rate:.2f};{c_rate:.2f};{u_rate:.2f};{u_mb:.0f};{c_mb:.3f};{c_ratio:.2f};{snr_ask:.0f};{snr_min:.2f};{snr_max:.2f};{snr_avg:.2f};{snr_median:.2f};{snr_total:.2f}"
    else:
        hdr_format = None
        data_format = """
Processed {brickcount} bricks of {bricksize} MB using {compressor} compressor
Elapsed (sec): read {r_time:6.2f} compress {c_time:6.2f} inflate {u_time:6.2f}
Rates  (MB/s): read {r_rate:6.2f} compress {c_rate:6.2f} inflate {u_rate:6.2f}
Compression ratio: {u_mb:.0f} MB / {c_mb:.3f} MB = {c_ratio:.2f}
Signal/Noise ratio: asked {snr_ask} min {snr_min:.2f} max {snr_max:.2f} avg {snr_avg:.2f} median {snr_median:.2f} overall {snr_total:.2f}"""

    if hdr_format: print(hdr_format)
    print(data_format.format(**info))

    if global_args.hist and global_args.hist[0]:
        print('Saving histogram as "{0}"'.format(global_args.hist[0]))
        import matplotlib.pyplot as plt
        #hist, bins = np.histogram(snrlist, bins=99, range=(0,99))
        plt.hist(snrlist, bins=99, range=(0,99))
        plt.savefig(global_args.hist[0])

if __name__ == "__main__":
    np.seterr(all='raise')
    parser = argparse.ArgumentParser(description='ZFP performance test.')
    parser.add_argument('filename', help='ZGY cube to operate on')
    parser.add_argument('--csv', action='store_true', help="Machine readable results")
    parser.add_argument('--csvheader', action='store_true', help="Machine readable results")
    parser.add_argument('--old', action='store_true', help="Use the old ZGY compressor instead of zfp")
    parser.add_argument('--sqnr', default=30, type=float, help='Target SQNR')
    parser.add_argument('--repeat', default=1, type=int, help='Run the decompress step this many times')
    parser.add_argument('--histogram', dest='hist', nargs=1, default="", type=str, help='Save histogram of actual SQNR values to png file')
    parser.add_argument('--buggy', action='store_true', help="Compressor expects C ordering")
    global_args = parser.parse_args()
    print(global_args)
    slurp(global_args.filename)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
Measure compression quality and compression factor.

Unlike test_zfp this does not try to measure performance.
The focus is to exercise higher levels of code (CompressPlugin)
and to be able to easily compare the effects of compression
parameters. Possibly by doing multiple compressions of each block.

Caveat: Since the code does both compress and decompress internally,
it might in principle "cheat" and get better results that way.
I have other tests that actually write out a compressed file and
reads it back in, and those seem to have similar results. So this
is probably ok.
"""

# Some things I would like to learn:
#
#   * Make a scatter plot of quality vs. compression and also the
#     compression- and decompression speed to evaluate which of ZFP
#     and the old ZGY compression is best.
#
#     Unsure whether I should use median or average, probably the latter.
#     Compare the old algorithm with ZFP. For ZFP I can also try
#     different strategies (precision vs. accuracy, int or float).
#     This currently requires tweaking the code in impl_zfp_compress;
#     maybe I'll add it as an argument to the compressor later.
#
#     Run several experiments, sort and load the results into a spreadsheet.
#     For each compression plug-in, plot the average or maybe the median
#     achieved snr agaist the achieved compression ratio.
#     Compare the plots for different algorithms to see which is better.
#     For a more accurate comparison it is possible to plot every
#     single brick's quality vs. compression. This requires some more code.
#     Possibly using matplotlib instead of relying on an external spreadsheet.
#     The reason I may want this extra detail is that neither the overall
#     (survey taken as a single brick) or average or median snr has a penalty
#     for having wildly varying quality between bricks. I could perhaps
#     measure e.g. the 20-percentile of the per block snr instead?
#     The most detailed information comes from plotting every single
#     data point.
#
#     May need to repeat this for several data sets, including several each
#     of float, int16, and int8 data. When preparing int cubes make sure
#     to clip outliers.
#
#     >>> Ongoing. In most cases Old ZGY compresses better but is slower.
#
#   * Measure the combined snr if a data set is first converted to int8
#     (or int16) and then compressed. Is the noise added in each step
#     independant?
#
#   * Accuracy of snr to zfp parameter heuristic.
#
#     To measure this I need to disable the goal seek algorithm.
#     Then run the tests and do a scatter plot of wanted vs. result snr.
#
#   * Which ZFP mode gives the best compression given an snr constraint?
#     And which is fastest for compress and fastest for decompress?
#
#       -- float data in "tolerance" mode [recommended in doc]
#       -- float data in "precision" mode
#       -- int data in "precision" mode [recommended in doc]
#       -- int data converted to float in "tolerance" mode
#       -- int data converted to float in "precision" mode
#
#      >>> Previous test results suggest the difference is not large.
#
#          The snr heuristic seems to work better in "precision" mode,
#          it might even be good enough to use this as our S/N metric
#          and skip the goal seek.
#
#          The two alternatives that convert int to float have an
#          additional caveat: If the app wants to read the data
#          as unscaled integers (to save memory) then there will be
#          an extra float -> int conversion that adds noise. So if
#          we choose any of those it might be better to lie when the
#          app asks for data type and claim it was float all along.
#
#   * Will ZFP give me better compression if I pass it 128^3 bricks?
#     >>> No. ZFP actually compresses each 4^3 block separately.
#
#   * Wish list: If the input is large, automatically crop out the
#     center and only analyze that. Or extract a random number of
#     brick-columns.

import numpy as np
import os, sys, math, time, argparse
from zgycompress import compress as old_compress_
from zgycompress import inflate as old_decompress_
from ..api import ZgyReader, ZgyWriter, ProgressWithDots, _map_SampleDataTypeToDataType
from ..impl.compress import CompressStats
from ..impl.zfp_compress import ZfpCompressPlugin
from ..impl import enum as impl_enum
from ..impl.bulk import ZgyInternalBulk
from ..iterator import readall
from ..test.utils import SDCredentials
from ..exception import *

global_args = None

def old_compress(data, sqnr):
    pad = np.array((64, 64, 64), dtype=np.int64) - np.array(data.shape)
    assert np.all(pad >= 0)
    if np.any(pad > 0):
        data = np.pad(data, ((0, pad[0]), (0, pad[1]), (0, pad[2])), 'edge')
    # The old compressor only handles float.
    data = data.astype(np.float32, copy=False)
    # Explicitly make the buffer fortran contiguous.
    # This is probably simpler than reversing the axis.
    data = np.copy(data, order="F")
    return old_compress_(data, sqnr)

def old_decompress(cdata):
    bricksize = (64, 64, 64) # Only works with this bricksize.
    tmp = old_decompress_(cdata, bricksize)
    return np.frombuffer(tmp, dtype=np.float32).reshape(bricksize, order="F")

class OldCompressPlugin:
    """
    Mock plugin, doesn't produce anything but does record the compression
    quality and compression rate for each brick.
    """
    @staticmethod
    def compress(idata, snr = 30, stats = None):
        # The old compressor only handles 64^3, so just log this as
        # being uncompressed. A brick size < 64^3 should work but we
        # end up padding every single brick and will probably get a
        # very bad compression rate.
        if np.any(np.array(idata.shape, dtype=np.int64) > 64):
            if stats:
                stats.add_data(idata, idata.size * idata.itemsize, idata)
                return
        starttime = time.perf_counter()
        cdata = old_compress(idata, snr)
        ctime = time.perf_counter() - starttime
        ddata = old_decompress(cdata)
        dtime = time.perf_counter() - starttime - ctime
        # Strip off any padding done on compress.
        # Note, this is a kludge. Normally we would have
        # the survey size available, so the old data need
        # not be present.
        s = idata.shape
        ddata = ddata[0:s[0], 0:s[1], 0:s[2]]
        if stats:
            stats.add_data(idata, len(cdata), ddata, ctime=ctime, dtime=dtime)

    def __init__(self, snr = 30):
        self._snr = snr
        self._details = "Old[target_snr={0:.1f}]".format(snr)
        self._stats = CompressStats(self._details)

    def __call__(self, data):
        return self.compress(data, self._snr, self._stats)

    def __str__(self):
        return self._details

    def dump(self, msg = None, *, outfile=None, text=True, csv=False, reset=True):
        self._stats.dump(msg, outfile=outfile, text=text, csv=csv)
        if reset:
            self._stats = CompressStats(details = self._details)

class CompressPluginPlain:
    """
    Mock plugin, doesn't produce anything but does record the compression
    quality for compression to 8 bit. Does not record timing statistics.
    """
    @staticmethod
    def compress(idata, wide = False, codingrange = (-1,1), stats = None):
        file_dtype = np.int16 if wide else np.int8
        if idata.dtype == np.float32:
            cdata = ZgyInternalBulk._scaleToStorage(idata, codingrange, file_dtype)
            ddata = ZgyInternalBulk._scaleToFloat(cdata, codingrange, file_dtype)
        elif idata.dtype == np.int8 or idata.dtype == file_dtype:
            cdata = idata.astype(file_dtype) # only needed for size
            ddata = idata # always perfect
        elif idata.dtype == np.int16 and file_dtype == np.int8:
            cdata = (idata // 256).astype(file_dtype)
            ddata = cdata.astype(idata.dtype) * 256
        else:
            assert False
        if stats:
            stats.add_data(idata, cdata.size * cdata.itemsize, ddata)
        global dejavu
    def __init__(self, codingrange = (-1,1), snr = 30):
        # TODO-Low, input arg "wide" instead of snr.
        self._codingrange = tuple(codingrange)
        self._wide = bool(snr <= 30)
        self._details = "int16" if self._wide else "int8"
        self._stats = CompressStats(self._details)

    def __call__(self, data):
        return self.compress(data, wide = self._wide, codingrange=self._codingrange, stats=self._stats)

    def __str__(self):
        return 'int16' if self._wide else 'int8'

    def dump(self, msg = None, *, outfile=None, text=True, csv=False, reset=True):
        self._stats.dump(msg, outfile=outfile, text=text, csv=csv)
        if reset:
            self._stats = CompressStats(details = self._details)

def find_center(r):
    """
    Crop partial bricks outside or below the survey since they
    would just confuse the statistics.
    If the file is large then crop out an area around the center.
    Try to read the full traces but limit the il and xl.
    """
    size = np.array(r.size, dtype=np.int64)
    cropsize = np.minimum(size, np.array([640,640,12800], dtype=np.int64))
    cropsize = (cropsize//64)*64
    cropoffset = ((size - cropsize)//128)*64
    if np.any(size - cropsize >= 64):
        print("Reading center", tuple(cropsize),
              "offset", tuple(cropoffset),
              "of survey size", tuple(r.size))
    return cropsize, cropoffset

def run_file(filename, logfilename):
    print("{0} -> {1}".format(filename, logfilename))
    with ZgyReader(filename, iocontext = SDCredentials()) as r:
        cropsize, cropoffset = find_center(r)
        compressors = []
        faclist = [ZfpCompressPlugin, OldCompressPlugin, CompressPluginPlain]
        snrlist = [ 20, 40, 30, 50, 10, 70, 60, ]
        snrlist.sort() # in batch mode, this is preferable.
        for factory in faclist:
            for snr in snrlist:
                compressors.append(factory(snr=snr))
                # Mega-kludge!
                if hasattr(compressors[-1], "_codingrange"):
                    compressors[-1]._codingrange = r.datarange

        processed = [0] * len(compressors)
        elapsed = [0] * len(compressors)
        progress = ProgressWithDots()
        chunksize = np.array((64, 64, 64), dtype=np.int64)
        # Set different from 64^3 to test different sizes in ZFP.
        # If selecting a block size > 64, the blocksize for reading
        # needs to be adjusted. Probably just set it to (N, N, 0).
        # Otherwise we might not get any full-chunksize bricks at all.
        # Similarly adjusted the cropping to avoid partial bricks.
        print("." * 51, file=sys.stderr, flush=True)
        # @@@ use dtype=np.float32 to convert everything to float first.
        for datastart, datasize, data in readall(r, chunksize=chunksize, progress=progress, maxbytes=128*1024*1024, cropsize=cropsize, cropoffset=cropoffset):
            if not np.all(datasize == chunksize):
                print("Oops: datasize", datasize, "chunksize", chunksize)
            assert np.all(datasize == chunksize) # Because file is cropped
            cix = 0
            for c in compressors:
                starttime = time.perf_counter()
                c(data)
                this_elapsed = time.perf_counter() - starttime
                elapsed[cix] += this_elapsed
                processed[cix] += data.size * data.itemsize
                cix += 1
                buffer_dtype = data.flat[0].dtype

        all_info = []
        with open(logfilename, "w") as csvfile:

            print(filename, file=csvfile)
            print("TABLE 1", file=csvfile)
            # This table measures a single time for compress and decompress
            # so it is not very useful except for validating table 2.
            # The overhead of goal seek is included both for old and new
            # algorithm.

            for cix in range(len(compressors)):
                if cix==0 or type(compressors[cix]) != type(compressors[cix-1]):
                    print("\n;Name;snr;size%;c+d MB/s;MB data;Elapsed", file=csvfile)
                s = compressors[cix]._stats # Gross violation of encapsulation.
                print(";{0};{1:.1f};{2:.1f};{3:.5g};{4:.5g};{5:.5g}".format(
                    str(compressors[cix]),
                    s.snr(),
                    100 * s._packed / max(1, s._original),
                    (processed[cix]/elapsed[cix])/(1024*1024),
                    processed[cix]/(1024*1024),
                    elapsed[cix]), file=csvfile)

            print("", file=csvfile)
            print("\nTABLE 2", file=csvfile)
            # This table measures compress- and decompress times separately.
            # TODO-Low: old algorithm includes goal seek overhead, while the
            # new one does not.

            for cix in range(len(compressors)):
                c = compressors[cix]
                if cix==0 or type(compressors[cix]) != type(compressors[cix-1]):
                    c.dump("", text=False, csv="header", reset=False, outfile=csvfile)
                all_info.append(c._stats._all_info)
                c.dump(str(c), text=False, csv=True, reset=False, outfile=csvfile)
                c.dump(str(c))

            if True:
                data = np.array(all_info, dtype=np.float32)
                data = data.reshape(len(faclist), len(snrlist), -1, 2)
                np.save(logfilename + ".npy", data)

            if False:
                s = data.shape
                data = data.reshape(s[0], s[1]*s[2], s[3])
                print(buffer_dtype, r.datatype.name, file=csvfile, sep="/")
                for c in faclist:
                    print(";SNR;{0} size%".format(str(c(snr=0))), file=csvfile, end="")
                print(file=csvfile)
                for pair in range(data.shape[1]):
                    for algo in range(data.shape[0]):
                        for value in range(data.shape[2]):
                            print(";{0:.1f}".format(data[algo,pair,value]), file=csvfile, end="")
                    print(file=csvfile)

if __name__ == "__main__":
    np.seterr(all='raise')
    parser = argparse.ArgumentParser(description='ZFP performance test.')
    parser.add_argument('filename', help='ZGY cube to operate on')
    parser.add_argument('--csv', default='TC.csv', help='Output easy to parse result')
    global_args = parser.parse_args()
    run_file(global_args.filename, global_args.csv)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



from .version import __version__
from ._accessor import SeisIO, SeisGeom, open_seisnc, coordinate_df

from ._seismic_dataset import (
    create_seismic_dataset,
    create3d_dataset,
    create2d_dataset,
)
